from lxml import etree
from typing import Dict
from os import path
from xml.sax.saxutils import XMLGenerator
import re
from sensestreet.deterministic_anonymization import anonimize_to_name, anonimize_to_hash
from functools import lru_cache
"""
BBXMLAnonymiser Module
This module provides functionality for anonymizing, and writing XML files while protecting sensitive information.

Classes:
- BBXMLAnonymiser: Handles the anonymization process of XML files.
- FakeUserGenerator: Generates anonymized user data.
Functions:
- anonymise_bbg_xml: A convenience function to anonymize an XML file.
- get_domain_from_email: Extracts the domain from an email address.
- anonimize_email: Creates an anonymized email using a given login name.
"""


class BBXMLAnonymiser:
    """
    BBXMLAnonymiser
    Handles the parsing, anonymizing, and writing of XML files.

    Attributes:
    - bank_pattern (str): A regex pattern used to identify company names related to a bank's role in conversations.
    - email_fields (set): XML tags representing email fields to anonymize.
    - anonimize_company (bool): Indicates if company-related information should be anonymized.
    - role_field (str or None): The XML tag used for identifying companies (optional).
    - bank_value (str): The replacement value for matching company names (default: "BANK").

    Methods:
    - anonymise_xml: Parses an XML file, anonymizes relevant fields, and writes the result to a new file.
    - _element_start: Processes the start of an XML element.
    - _element_end: Processes the end of an XML element and performs anonymization as needed.
    """
    def __init__(
        self,
        bank_pattern,
        role_field=None,
        bank_value="BANK",
    ):
        self.bank_pattern = bank_pattern
        self.email_fields = {"EmailAddress", "CorporateEmailAddress"}
        self.anonimize_company = role_field is not None
        self.role_field = role_field
        self.bank_value = bank_value
        self._xmlwriter = None
        self._current_user = None
        self._inside_user = False
        self.user_generator = FakeUserGenerator(anonimize_company=self.anonimize_company)

    def anonymise_xml(self, xml_in, xml_out):
        context = etree.iterparse(xml_in, events=("start", "end"))
        self._inside_user = False
        with open(xml_out, "w", encoding="UTF-8") as fp:
            self._xmlwriter = XMLGenerator(
                fp, encoding="UTF-8", short_empty_elements=False
            )
            self._xmlwriter.startDocument()

            for event, elem in context:
                if event == "start":
                    self._element_start(elem)
                elif event == "end":
                    self._element_end(elem)
                    elem.clear()

            self._xmlwriter.endDocument()

    def _element_start(self, elem):
        self._xmlwriter.startElement(name=elem.tag, attrs=elem.attrib)
        if elem.tag == "User":
            self._inside_user = True

        if self._inside_user and elem.tag == "LoginName" and elem.text:
            self._current_user = self.user_generator.generate_user_data(elem.text)

    def _element_end(self, elem):
        if elem.tag == "LoginName" and elem.text:
            self._current_user = self.user_generator.generate_user_data(elem.text)

        if self.role_field and elem.tag == self.role_field and re.search(
            self.bank_pattern, str(elem.text), flags=re.IGNORECASE
        ):
            elem.text = self.bank_value
            if self._inside_user:
                self._current_user[elem.tag] = self.bank_value
        elif (
            self._inside_user and elem.tag in self.user_generator.elements_to_anonymise
        ):
            elem.text = self._current_user[elem.tag]
        elif (
            self._inside_user and elem.tag in self.email_fields and elem.text is not None
        ):
            elem.text = anonimize_email(elem.text, self._current_user['LoginName'])

        if elem.text:
            self._xmlwriter.characters(elem.text)

        self._xmlwriter.endElement(name=elem.tag)
        self._xmlwriter.characters("\n")

        if elem.tag == "User":
            self._inside_user = False
            self._current_user = None

def get_domain_from_email(email):
    try:
        domain = email.split('@')[1]
        return domain
    except IndexError:
        return ""
    
def anonimize_email(email, login):
    return f"{login}@{get_domain_from_email(email)}"
    
class FakeUserGenerator:
    def __init__(self, anonimize_company=False):
        self.elements_to_anonymise = {
            "LoginName",
            "FirstName",
            "LastName",
            "UUID",
            "FirmNumber",
            "AccountNumber",
        }
        self.anonimize_company = anonimize_company

        if self.anonimize_company:
            self.elements_to_anonymise.add("CompanyName")
            self.elements_to_anonymise.add("EmailAddress")
            self.elements_to_anonymise.add("CorporateEmailAddress")

    @lru_cache(maxsize=100)
    def generate_user_data(self, login) -> Dict:
        first_name, last_name = anonimize_to_name(login)
        login = anonimize_to_hash(login, short=False)

        anonimized_data = {
            "FirstName": first_name,
            "LastName": last_name,
            "LoginName": login,
            "UUID": "",
            "FirmNumber": "",
            "AccountNumber": "",
        }
    
        if self.anonimize_company:
            anonimized_data["CompanyName"] = "ACME"
            anonimized_data["CorporateEmailAddress"] = f"{login}@acme.com"
            anonimized_data["EmailAddress"] = f"{login}@acme.com"

        return anonimized_data
            

def anonymise_bbg_xml(xml_in, xml_out, bank_pattern=None):
    """
    A convenience function to create an instance of BBXMLAnonymiser and anonymize an XML file.

    Parameters:
    - xml_in (str): Input XML file path.
    - xml_out (str): Output file path where anonymized XML will be saved.
    - bank_pattern (str): Regular expression pattern to identify banks in XML data.
    """
    anonymiser = BBXMLAnonymiser(bank_pattern=bank_pattern)
    anonymiser.anonymise_xml(xml_in, xml_out)


if __name__ == "__main__":
    anonymise_bbg_xml(
        "./example.xml",
        "./test.xml"
    )



LAST_NAMES = [
    "Chi",
    "Burgett",
    "Schuetz",
    "Rowe",
    "Haddox",
    "Tercero",
    "Reeser",
    "Oakman",
    "Ledner",
    "Arroyo",
    "Raynor",
    "Shoup",
    "Brandenberger",
    "Rosenbaum",
    "Armitage",
    "Neilsen",
    "Taube",
    "Stehr",
    "Hoeger",
    "Mcgowan",
    "Carns",
    "Meaders",
    "Montag",
    "Beckham",
    "Castello",
    "Conn",
    "Thompson",
    "Right",
    "Rand",
    "Farias",
    "Hagenes",
    "Volkmann",
    "Wilkinson",
    "Robel",
    "Scotto",
    "Gislason",
    "Faunce",
    "Gurr",
    "May",
    "Raminez",
    "Dinh",
    "Konieczny",
    "Dibella",
    "Thomas",
    "Fritsch",
    "Almonte",
    "Shoaf",
    "West",
    "Nader",
    "Rotondo",
    "Ziemann",
    "Junkin",
    "Zboncak",
    "Frami",
    "Bold",
    "Lucarelli",
    "Roberts",
    "Stuck",
    "Pavon",
    "Spencer",
    "Orndorff",
    "Beaulieu",
    "Ormsby",
    "Montoya",
    "Canales",
    "Bichrest",
    "Eide",
    "Enciso",
    "Bemis",
    "Liechty",
    "Fettig",
    "Grady",
    "King",
    "Renken",
    "Marcus",
    "Champlin",
    "Ericksen",
    "Coil",
    "Lewandowski",
    "Judd",
    "Huggett",
    "Cantor",
    "Seawood",
    "Border",
    "Albee",
    "Beatty",
    "Tocci",
    "Holgate",
    "Hirthe",
    "Marro",
    "Lubowitz",
    "Malloy",
    "Mraz",
    "Aikins",
    "Stern",
    "Drewry",
    "Mcandrews",
    "Ames",
    "Coryell",
    "Kohler",
    "Lockman",
    "Swanson",
    "Selvy",
    "Pinedo",
    "Garnes",
    "Peery",
    "Omalley",
    "Halliwell",
    "Carter",
    "Leeks",
    "Reinger",
    "Woolum",
    "Murdoch",
    "Smidt",
    "Hammes",
    "Schroeder",
    "Kuphal",
    "Huddle",
    "Kaspar",
    "Jimison",
    "O'Connell",
    "Mcfarland",
    "Liston",
    "Rice",
    "Tindel",
    "Chae",
    "Trojacek",
    "Mastro",
    "Balog",
    "Kanagy",
    "Hiemstra",
    "Melanson",
    "Genovese",
    "Leonhardt",
    "Rakestraw",
    "Diem",
    "Kistner",
    "Halvorson",
    "Work",
    "Kimberling",
    "Jast",
    "Pellegrino",
    "Voisine",
    "Maher",
    "Shanahan",
    "Melody",
    "Styers",
    "Hagar",
    "McDermott",
    "Haag",
    "Whittier",
    "Wiltshire",
    "Lipton",
    "Abel",
    "Hopf",
    "Romaguera",
    "Glavin",
    "Balboa",
    "Gamache",
    "Zamorano",
    "Towne",
    "Millican",
    "Rau",
    "Esses",
    "Thorp",
    "Averitt",
    "Siegfried",
    "Starr",
    "Holliman",
    "Straw",
    "Wold",
    "Kirkham",
    "Halfacre",
    "Anchondo",
    "Borel",
    "Brasfield",
    "Kling",
    "Wichman",
    "Ashford",
    "Melendez",
    "Trottier",
    "Lightford",
    "Voshell",
    "Volpe",
    "Maine",
    "Volz",
    "Jasper",
    "Lavoie",
    "Oller",
    "Twiggs",
    "Krieg",
    "Schuppe",
    "Pesqueira",
    "Fike",
    "Kandoll",
    "Seegmiller",
    "Tootle",
    "Callaghan",
    "Gatto",
    "Shook",
    "Hafer",
    "Lingenfelter",
    "Mosciski",
    "Mattis",
    "Weimann",
    "Polen",
    "Labombard",
    "Gard",
    "Keltz",
    "Bashirian",
    "Bogisich",
    "Thiel",
    "Armes",
    "Koepp",
    "Rio",
    "Diangelo",
    "Campo",
    "Hock",
    "Padua",
    "Sypher",
    "Piekarski",
    "Melancon",
    "Deshazo",
    "Stallings",
    "Coddington",
    "Eichmann",
    "Erdman",
    "Gossard",
    "Wilczynski",
    "Angeles",
    "Payan",
    "Absher",
    "Brugger",
    "Albertini",
    "Gargano",
    "Mante",
    "Kennemer",
    "Goya",
    "Goldner",
    "Jerome",
    "Sentell",
    "Senger",
    "Adkins",
    "Magers",
    "Backer",
    "Duda",
    "Ohair",
    "Seamon",
    "Woolfolk",
    "Reynolds",
    "Farrish",
    "Chadburn",
    "Morman",
    "Rockhill",
    "Mcfarlin",
    "Fannin",
    "Bickel",
    "Gallman",
    "Rudloff",
    "Chmiel",
    "Mijangos",
    "Whitten",
    "Mccaa",
    "Muller",
    "Seabrooks",
    "Kub",
    "Lueilwitz",
    "Jiggetts",
    "Wyman",
    "Kolbe",
    "Loos",
    "Whitcomb",
    "Bowlin",
    "Bibbs",
    "Calender",
    "Kassulke",
    "Rolfson",
    "Weisman",
    "Zinck",
    "Bramblett",
    "Breneman",
    "Emmons",
    "Holly",
    "Ryan",
    "Mckinzie",
    "Lemley",
    "Hempstead",
    "Ohlinger",
    "Policastro",
    "Mertz",
    "Windler",
    "Batz",
    "Mcgrew",
    "Schlottmann",
    "Engstrom",
    "Degreenia",
    "Gulgowski",
    "Afanador",
    "Holz",
    "Bittner",
    "Mccaster",
    "Trollinger",
    "Lamantia",
    "Clinard",
    "Conner",
    "Baumbach",
    "Marshburn",
    "Colligan",
    "Resto",
    "Dambrosia",
    "Kalina",
    "Tremblay",
    "Funderburg",
    "Douglas",
    "Dundon",
    "Beers",
    "Decola",
    "Rikard",
    "Ferry",
    "Parker",
    "Bischoff",
    "Creamer",
    "Cartier",
    "Pagac",
    "Villela",
    "Kemmer",
    "Lokken",
    "Ohern",
    "Deyoung",
    "Bettcher",
    "Bethke",
    "Hinojos",
    "Laclair",
    "Alling",
    "Press",
    "Krawczyk",
    "Simmers",
    "Wilmes",
    "Kranz",
    "Burdette",
    "Tietjen",
    "Moss",
    "Pfannerstill",
    "Jennings",
    "Dickison",
    "Bode",
    "Crona",
    "Rakowski",
    "Kinsey",
    "O'Conner",
    "Terry",
    "Lone",
    "Labadie",
    "Canup",
    "Kreger",
    "Graffam",
    "Wiltse",
    "Spigner",
    "Propp",
    "Gleason",
    "Saine",
    "Cleghorn",
    "Andrzejewski",
    "Falgoust",
    "Elem",
    "Thiele",
    "Edmiston",
    "Santistevan",
    "Eddleman",
    "Bahringer",
    "Lebron",
    "Diller",
    "Bullard",
    "Kilback",
    "Lentine",
    "Ligon",
    "China",
    "Espino",
    "Dudgeon",
    "Malchow",
    "Kuehner",
    "Jerde",
    "Bittinger",
    "Hakes",
    "Lindley",
    "Dyment",
    "Bonney",
    "Dyess",
    "Carley",
    "Constable",
    "Whistler",
    "Schowalter",
    "Motz",
    "Coria",
    "Aasen",
    "Wrinkle",
    "Brunswick",
    "Garibaldi",
    "Steger",
    "Batdorf",
    "Shen",
    "Bouska",
    "Stoltenberg",
    "Hoyte",
    "Tallon",
    "Lamphere",
    "Daugherty",
    "Weir",
    "Cane",
    "Froman",
    "Aldinger",
    "Lemons",
    "Luciani",
    "Hightower",
    "Fogarty",
    "Sisler",
    "Jenney",
    "Lessman",
    "Munsterman",
    "Rome",
    "Kendig",
    "Batres",
    "Mccleery",
    "Flatley",
    "Binder",
    "Showers",
    "Bushway",
    "Allinder",
    "Whitby",
    "Larson",
    "Dicki",
    "Farabaugh",
    "Brigmond",
    "Borey",
    "Bufford",
    "Covert",
    "Sheilds",
    "Philbrick",
    "Green",
    "Jaworski",
    "Stroman",
    "Terrell",
    "Plascencia",
    "Zieman",
    "Lampman",
    "Carroll",
    "Hatcher",
    "Leisy",
    "Wilderman",
    "Jaskolski",
    "Kuhic",
    "Leclerc",
    "Kunde",
    "Rempel",
    "Crofoot",
    "Steinert",
    "Sun",
    "Gravelle",
    "Benito",
    "Vaughn",
    "Clevenger",
    "Wark",
    "Farrell",
    "Kreiger",
    "Kirlin",
    "Bakker",
    "Arent",
    "Weiler",
    "Elfrink",
    "Dooley",
    "Smitham",
    "Heatherly",
    "Devries",
    "Bertucci",
    "Haan",
    "Imel",
    "Farris",
    "Sarabia",
    "Stackpole",
    "Emard",
    "Stebbins",
    "Lindbloom",
    "Macrae",
    "Jilek",
    "Emigh",
    "McKenzie",
    "Mccrae",
    "Bartell",
    "Orman",
    "Greenburg",
    "Cybulski",
    "Slevin",
    "Biros",
    "Lilienthal",
    "Clepper",
    "Quigley",
    "Wurtz",
    "Shahid",
    "Marquardt",
    "Gebo",
    "Loveless",
    "Lajeunesse",
    "Paolucci",
    "Badgett",
    "Deschamp",
    "Klemme",
    "Ortiz",
    "Clyburn",
    "Wierenga",
    "Girard",
    "Mcpeters",
    "Rieke",
    "Bender",
    "Romney",
    "Luth",
    "Pardo",
    "Kruger",
    "Spies",
    "Nelligan",
    "Backus",
    "Ranney",
    "Lecompte",
    "Bow",
    "Guilliams",
    "Jacobson",
    "Bergstrom",
    "Wetherell",
    "Lefevers",
    "Julian",
    "Gottlieb",
    "Watrous",
    "Stephens",
    "Block",
    "Hauff",
    "Linen",
    "Godfrey",
    "Dinkins",
    "Trudell",
    "Wehner",
    "Cockrum",
    "Lavelle",
    "Hemminger",
    "Bednar",
    "Archibald",
    "Primer",
    "Constance",
    "Custis",
    "Liz",
    "Bagnall",
    "Dahmer",
    "Turberville",
    "Nell",
    "Clow",
    "Hollenbach",
    "Sines",
    "Blick",
    "Franecki",
    "Troyer",
    "Gerlach",
    "Davies",
    "Wiza",
    "Kohr",
    "Ogan",
    "Nalls",
    "Watchman",
    "Alverson",
    "Goodwin",
    "Dewar",
    "Swope",
    "Joshi",
    "Burger",
    "Biederman",
    "Ruse",
    "Buckridge",
    "Cephus",
    "Borror",
    "Risk",
    "Hutton",
    "Reynosa",
    "Pfeil",
    "Wunsch",
    "Wehmeyer",
    "Mcmeen",
    "Saliba",
    "Heldt",
    "Armistead",
    "Sharpton",
    "Berger",
    "Yetter",
    "Dalzell",
    "Barfoot",
    "Bulfer",
    "Wronski",
    "Craner",
    "Hufford",
    "Muncy",
    "Schaden",
    "Pett",
    "Hanneman",
    "Purdy",
    "Townson",
    "Magill",
    "Yundt",
    "Augusta",
    "Pare",
    "Troester",
    "Saeed",
    "Klann",
    "Martensen",
    "Murry",
    "Mcmillin",
    "Leeds",
    "Lou",
    "Strobel",
    "Swan",
    "Tome",
    "Lucca",
    "Ferris",
    "Schreiber",
    "Rosecrans",
    "Tadlock",
    "Loggins",
    "Dickinson",
    "Siers",
    "Jacobs",
    "Mars",
    "Schmeler",
    "Robbs",
    "Crowther",
    "Kuhlman",
    "Gust",
    "Delaune",
    "Marley",
    "Wolken",
    "Barrientos",
    "Villalpando",
    "Agron",
    "Stoute",
    "Freese",
    "Ogren",
    "Glazer",
    "Downs",
    "Nikolaus",
    "Sherlock",
    "Bethel",
    "Mcspadden",
    "Verrett",
    "Hauck",
    "Terhaar",
    "Sheahan",
    "Granado",
    "Keep",
    "Ciesla",
    "Client",
    "Neihoff",
    "Shores",
    "Goodloe",
    "Romaine",
    "Mcmanis",
    "Mast",
    "Fritz",
    "Herrod",
    "Bernhardt",
    "Skiles",
    "Keebler",
    "Robards",
    "Greely",
    "Renner",
    "Conroy",
    "Hardman",
    "Dach",
    "Eden",
    "Corbell",
    "Au",
    "Weissnat",
    "Reasoner",
    "Clifton",
    "Wanamaker",
    "Ankrom",
    "Bingham",
    "Renteria",
    "Caines",
    "Jakubowski",
    "Millener",
    "Cimino",
    "Venema",
    "Metheny",
    "Nurse",
    "Defore",
    "Faria",
    "Goolsby",
    "Deford",
    "Reichel",
    "Cappiello",
    "Tierney",
    "Pitre",
    "Dresel",
    "Lufkin",
    "Ritchie",
    "Buell",
    "Lauber",
    "Brungardt",
    "Murazik",
    "Nitta",
    "Hanna",
    "Friesen",
    "Devlin",
    "Como",
    "Lares",
    "Collier",
    "Westman",
    "Homan",
    "Gonzalas",
    "Andreotti",
    "Fitzhugh",
    "Freitag",
    "Troiano",
    "Broderson",
    "Miskell",
    "Ratliff",
    "Gullick",
    "Toller",
    "Vann",
    "Zych",
    "Thomsen",
    "Couture",
    "Hartmann",
    "Cha",
    "Titcomb",
    "Miyamoto",
    "Hoppe",
    "Eisenman",
    "Hasler",
    "Meneses",
    "Gagnier",
    "Murton",
    "Swing",
    "Vandervort",
    "Zuniga",
    "Chadwell",
    "Abadie",
    "Heathcote",
    "Knutsen",
    "Stronach",
    "Addington",
    "Sporer",
    "Langworth",
    "Liebel",
    "Neace",
    "Streight",
    "Tow",
    "Calderwood",
    "Bevel",
    "Flanery",
    "Harbin",
    "Andino",
    "Metz",
    "Jeffreys",
    "Bella",
    "Stark",
    "Sciacca",
    "Murray",
    "Beer",
    "Escamilla",
    "Pitcher",
    "Carollo",
    "Rayburn",
    "Southwick",
    "Dao",
    "Schmitt",
    "Murphy",
    "Gilford",
    "Molinar",
    "Groff",
    "Miller",
    "Sipes",
    "Bergnaum",
    "Delgiudice",
    "Youngquist",
    "Godbout",
    "Swiger",
    "Kinch",
    "Poquette",
    "Barlow",
    "Bays",
    "Wolff",
    "Sauer",
    "Mebane",
    "Owings",
    "Klein",
    "Pietrzak",
    "Edick",
    "Acey",
    "Heaney",
    "Newbrough",
    "Kertzmann",
    "Davey",
    "Bradham",
    "Pasko",
    "Campana",
    "Hettinger",
    "Bogan",
    "Abshire",
    "Rosenblatt",
    "Brehmer",
    "Laliberte",
    "Summerfield",
    "Baillargeon",
    "Bailey",
    "Mika",
    "Ketcham",
    "Dees",
    "Knebel",
    "Bialek",
    "Rickman",
    "Wasko",
    "Thomson",
    "Heidenreich",
    "Aubert",
    "Rizzo",
    "Cage",
    "Rojo",
    "Mulvihill",
    "Bins",
    "Lebsock",
    "Hane",
    "Hilpert",
    "Moreles",
    "Boddie",
    "Looney",
    "Kealey",
    "Feehan",
    "Gosier",
    "Browder",
    "Davis",
    "Scull",
    "Gasaway",
    "Ankunding",
    "Mcguire",
    "Mitchell",
    "Vanderhoof",
    "Schmidt",
    "Mattinson",
    "Lanford",
    "Solt",
    "Postma",
    "Graziano",
    "Roger",
    "Lanier",
    "Merriam",
    "Lukes",
    "Nitzsche",
    "Stoecker",
    "Studivant",
    "Harlow",
    "Ring",
    "Lovitt",
    "Volkman",
    "Rutherford",
    "Sikes",
    "Cremin",
    "Pavlik",
    "Amerine",
    "Herder",
    "Lenhardt",
    "Scudder",
    "Melnyk",
    "Vasquez",
    "Veum",
    "Emmerich",
    "Badalamenti",
    "Vanderford",
    "Howell",
    "Puente",
    "Sward",
    "Alameda",
    "Mcentire",
    "Labelle",
    "Wilburn",
    "Frandsen",
    "Flannigan",
    "Adcock",
    "Simonis",
    "Thakkar",
    "Milne",
    "Spidle",
    "Freeborn",
    "Guarino",
    "Smith",
    "Harvard",
    "Schaefer",
    "Tarr",
    "Guider",
    "Zulauf",
    "Borer",
    "Strother",
    "Haydon",
    "Burkhardt",
    "Music",
    "Kiehn",
    "Ferrier",
    "Durgan",
    "Camacho",
    "Olson",
    "Galyean",
    "Letsinger",
    "Leitner",
    "Reagle",
    "Dickens",
    "Arboleda",
    "Litz",
    "Boomhower",
    "Heather",
    "Spalla",
    "Seltzer",
    "Stracke",
    "Sartain",
    "Villicana",
    "Moone",
    "Master",
    "Wynter",
    "Pfeffer",
    "Depaolo",
    "Neal",
    "Duhart",
    "Sabbagh",
    "Guidroz",
    "Cummerata",
    "Heindel",
    "Lover",
    "Markley",
    "Moulton",
    "Perla",
    "Cutler",
    "Vanalstyne",
    "Wisoky",
    "Kaiser",
    "Bouie",
    "Roepke",
    "Delp",
    "Hubbs",
    "McCullough",
    "Levan",
    "Mayers",
    "Magwood",
    "Eisele",
    "Raven",
    "Pio",
    "Days",
    "Antoine",
    "Seeber",
    "Hansen",
    "Monsen",
    "Tosi",
    "Hoffman",
    "Tallmadge",
    "Hsieh",
    "Larose",
    "Johanson",
    "Wolf",
    "Slaton",
    "Riter",
    "Flagg",
    "Pfannenstiel",
    "Neiman",
    "Armbruster",
    "Chastain",
    "Monahan",
    "Dumlao",
    "Berge",
    "Ragin",
    "Bauch",
    "Paucek",
    "Watts",
    "Ruecker",
    "Needles",
    "Satcher",
    "Wymore",
    "Sabol",
    "Wind",
    "Stupka",
    "Saylors",
    "Mazzoni",
    "Furness",
    "Smiley",
    "Lloyd",
    "Proto",
    "Medeiros",
    "Lundy",
    "Klocko",
    "Warwick",
    "Vanepps",
    "Mccluney",
    "Lasso",
    "Dangelo",
    "Wert",
    "Wain",
    "Winton",
    "Perryman",
    "Kosak",
    "Runolfsson",
    "Padberg",
    "Grant",
    "Sharlow",
    "Applebee",
    "Sturdivant",
    "Feil",
    "Biffle",
    "Kimbrell",
    "Modeste",
    "Bing",
    "Sather",
    "Twigg",
    "Pettiford",
    "Little",
    "Deyoe",
    "Kmetz",
    "Bosco",
    "Schuster",
    "Luettgen",
    "Castilleja",
    "Wimbish",
    "Books",
    "Varona",
    "Maybee",
    "Crist",
    "Miser",
    "Wyse",
    "Fay",
    "Prohaska",
    "Mehr",
    "O'Hara",
    "Fleig",
    "Luther",
    "Winner",
    "Furey",
    "Tucci",
    "Daluz",
    "Waller",
    "Jhonson",
    "Warman",
    "Vanatta",
    "Dearborn",
    "Ebert",
    "Bruss",
    "Nicolas",
    "Parra",
    "Morais",
    "Lopez",
    "Dodson",
    "Treadwell",
    "Bartle",
    "Tijerina",
    "Authement",
    "Benjamin",
    "Weinstock",
    "Goyette",
    "Stillings",
    "Braun",
    "Jewell",
    "River",
    "Than",
    "Seaver",
    "Dupler",
    "Sillman",
    "Nadler",
    "Mccarron",
    "Streich",
    "Cuccia",
    "Gutmann",
    "Overbey",
    "Galgano",
    "Oberbrunner",
    "Mckeown",
    "Wicker",
    "Raybon",
    "Hanks",
    "Abdalla",
    "Steffens",
    "Werth",
    "Bouley",
    "Brayman",
    "Chamberlain",
    "Musgrove",
    "Maughan",
    "Meadow",
    "Thetford",
    "Huddleston",
    "Dively",
    "Roberie",
    "Mueller",
    "Descoteaux",
    "Rotolo",
    "Hauser",
    "Knoll",
    "Kerluke",
    "Hodges",
    "Krueger",
    "Mcmillan",
    "Gobeil",
    "Frink",
    "Chambers",
    "Dry",
    "Bhakta",
    "Boyce",
    "Hermann",
    "Racey",
    "Bourgault",
    "Fortney",
    "Kleven",
    "Carraway",
    "Hamlin",
    "Quam",
    "Keo",
    "Bradley",
    "Vince",
    "Mullan",
    "Kropf",
    "Calderone",
    "Lang",
    "Hemmings",
    "Nienow",
    "Dolph",
    "Fail",
    "Claybrook",
    "Mcmunn",
    "Boden",
    "Wanner",
    "Earles",
    "Devaney",
    "Fausto",
    "Trotta",
    "Dolson",
    "Johnson",
    "Navarette",
    "Dantonio",
    "Foushee",
    "Tubman",
    "Tromp",
    "Redd",
    "Adams",
    "Ferber",
    "Clingman",
    "Forsman",
    "Mann",
    "Piermarini",
    "Bradtke",
    "Dibbert",
    "Weeks",
    "Scherrer",
    "Shealy",
    "Fiedler",
    "Edgar",
    "Talley",
    "Schimmel",
    "Noles",
    "Ravenscroft",
    "Gelinas",
    "Skillman",
    "Kshlerin",
    "Sackett",
    "Kos",
    "Pecora",
    "Rath",
    "Gusikowski",
    "Nicastro",
    "Postel",
    "Necessary",
    "Medders",
    "Cratty",
    "Stepney",
    "Hartgrove",
    "Mellon",
    "Kellen",
    "Sawayn",
    "Files",
    "Alvis",
    "Mccumber",
    "Torphy",
    "Vickers",
    "Skipworth",
    "Davilla",
    "Markland",
    "Chitty",
    "Janes",
    "Kuchta",
    "Llewellyn",
    "Bayne",
    "Rippin",
    "Walsh",
    "Gutkowski",
    "Burfield",
    "Snelling",
    "Hargett",
    "Loredo",
    "Farner",
    "Riviera",
    "Freund",
    "Shoultz",
    "Dale",
    "Fultz",
    "Seeger",
    "Copland",
    "Branum",
    "Ruder",
    "Bernhard",
    "Creighton",
    "Mazurek",
    "Houseknecht",
    "Bjornson",
    "Ryals",
    "Ocon",
    "Dietrich",
    "Hills",
    "Lynes",
    "Warfield",
    "Tiller",
    "Cleary",
    "Okeefe",
    "Torrence",
    "Durante",
    "Prokop",
    "Vitolo",
    "Millard",
    "Wisozk",
    "Yocum",
    "Kelton",
    "Conlan",
    "Hanshaw",
    "Sevilla",
    "Dudding",
    "Bostic",
    "Achenbach",
    "Christiansen",
    "Rahman",
    "Klopfer",
    "Benningfield",
    "Virden",
    "Kozey",
    "Calton",
    "Akey",
    "Stokes",
    "Blazek",
    "Brandel",
    "Hyppolite",
    "Marks",
    "Swaniawski",
    "Crimmins",
    "Wiegand",
    "Cobian",
    "Arnott",
    "Varnell",
    "Rothe",
    "Chattin",
    "Goetsch",
    "Mcpherson",
    "Bumbrey",
    "Eshleman",
    "Wideman",
    "Kirwin",
    "Machin",
    "Holt",
    "Mcpeak",
    "Wintheiser",
    "Wigington",
    "Ballenger",
    "Allsup",
    "Cranfield",
    "Kessler",
    "Fleischer",
    "Belford",
    "Gately",
    "Hegmann",
    "Guilbert",
    "Yocom",
    "Rubalcaba",
    "Splawn",
    "Horan",
    "Goodson",
    "Temme",
    "Blind",
    "Twilley",
    "Hochmuth",
    "Zamarripa",
    "Sarmiento",
    "Neisler",
    "Placencia",
    "Cartwright",
    "Armstrong",
    "Adcox",
    "Feeney",
    "Chung",
    "Salaam",
    "Briley",
    "Near",
    "Hobart",
    "Maciel",
    "Hardesty",
    "Helmers",
    "Starnes",
    "Renfroe",
    "Barberio",
    "Wiser",
    "Rogahn",
    "Eutsler",
    "Maly",
    "Clendenin",
    "Hartig",
    "Plante",
    "Stolte",
    "Heyen",
    "Bobby",
    "Linley",
    "Chartrand",
    "Feltz",
    "Coyle",
    "Teamer",
    "Schiller",
    "Greenholt",
    "Will",
    "Schulist",
    "Granillo",
    "Delillo",
    "Trantow",
    "Wu",
    "Hanger",
    "Bean",
    "Pitcock",
    "Staton",
    "Conyers",
    "Due",
    "Mchargue",
    "Stoller",
    "Amburn",
    "Mills",
    "Kump",
    "Cartledge",
    "Bober",
    "Isenberg",
    "Lippert",
    "Tony",
    "Wilner",
    "Koffler",
    "Medal",
    "Verne",
    "Allman",
    "Bodiford",
    "Jarnigan",
    "Bruneau",
    "Becker",
    "Beverley",
    "Hagood",
    "Blish",
    "Kapoor",
    "Villalta",
    "Schoen",
    "Yarber",
    "Briganti",
    "Saur",
    "Croff",
    "Corkill",
    "Keitt",
    "Capito",
    "White",
    "Huling",
    "Harvey",
    "Fripp",
    "Lehner",
    "Redner",
    "Lemoine",
    "Braithwaite",
    "Arviso",
    "Riles",
    "Moats",
    "Honea",
    "Patt",
    "Simpson",
    "Gourlay",
    "Fahey",
    "Barrows",
    "Luera",
    "Noonkester",
    "Reilley",
    "Grimes",
    "Olszewski",
    "Tacy",
    "Zielinski",
    "Theriot",
    "Nocera",
    "Rink",
    "Reitzel",
    "Lindgren",
    "Cousin",
    "Spicher",
    "Coffelt",
    "Alexis",
    "Hoerr",
    "Veliz",
    "Roda",
    "Slaugh",
    "Stancill",
    "Rodenberger",
    "Olivier",
    "Garrels",
    "Rippe",
    "Younkin",
    "Vales",
    "Fierros",
    "Laughridge",
    "Gleichner",
    "McClure",
    "Denker",
    "Glasgow",
    "Colella",
    "Harrington",
    "Katzman",
    "Leblanc",
    "Cun",
    "Shiner",
    "Felton",
    "Rayo",
    "Lawry",
    "Agnew",
    "Richard",
    "Phinney",
    "Bridge",
    "Isenhour",
    "Myers",
    "Moret",
    "Kirsch",
    "Razor",
    "Pacocha",
    "Jones",
    "Chabot",
    "Gillispie",
    "Nave",
    "Foland",
    "Harrelson",
    "Siemers",
    "Carnell",
    "Damron",
    "Lefever",
    "Danziger",
    "Bolles",
    "Newbold",
    "Torgeson",
    "Toy",
    "Doyle",
    "Howe",
    "Shields",
    "Acheson",
    "Spinelli",
    "Rosol",
    "Seth",
    "Gillett",
    "Mistretta",
    "Kuhn",
    "Gerhold",
    "Kline",
    "Dare",
    "Poehler",
    "Ponder",
    "Fenwick",
    "Scalf",
    "Kwong",
    "Capote",
    "Bechtelar",
    "Heineman",
    "Mayert",
    "Mcduffy",
    "Mccready",
    "Klopp",
    "Cole",
    "Jeffords",
    "Sorrels",
    "Voltz",
    "Kaestner",
    "Ing",
    "Muse",
    "Licht",
    "Doris",
    "Koss",
    "Degraffenreid",
    "Goodpaster",
    "Goodwill",
    "Rohan",
    "Resch",
    "Macejkovic",
    "Duty",
    "Patrick",
    "Ziebarth",
    "Grajeda",
    "Bartley",
    "Gorczany",
    "Almon",
    "Liedtke",
    "Sirianni",
    "Toledo",
    "Harsh",
    "Mohr",
    "Hamill",
    "Majeed",
    "Tharpe",
    "Schuman",
    "Wolter",
    "Tullius",
    "Reichert",
    "Tiemann",
    "Revilla",
    "Izaguirre",
    "Alanis",
    "Mooring",
    "Russel",
    "Lofthouse",
    "Dutra",
    "Harty",
    "Devita",
    "Ortego",
    "Brubaker",
    "Boudreau",
    "Coleman",
    "Raglin",
    "Taing",
    "Leedy",
    "Schneider",
    "Price",
    "Grassi",
    "Kennard",
    "Munday",
    "Vidrio",
    "Howey",
    "Sae",
    "Kuhlmann",
    "Ware",
    "Verde",
    "Kovacek",
    "Johnsrud",
    "Curry",
    "Ullrich",
    "Mable",
    "Lynch",
    "Streicher",
    "Lubinski",
    "Hughley",
    "Heier",
    "Chamberlin",
    "Tell",
    "Zemlak",
    "Dewolf",
    "Richart",
    "Syed",
    "Caudill",
    "Lesniak",
    "Zieme",
    "Reinbold",
    "Tokarski",
    "Tillman",
    "Brawner",
    "Wroten",
    "Culligan",
    "Basel",
    "Zeiler",
    "Hults",
    "Bartoletti",
    "Jenkins",
    "Knipe",
    "Lablanc",
    "Cadieux",
    "Littel",
    "Buchholtz",
    "Para",
    "Jolly",
    "Gleaton",
    "Rotter",
    "Mortimore",
    "Ashburn",
    "Grove",
    "Johns",
    "Mullenax",
    "Wyrick",
    "Chynoweth",
    "Burkley",
    "Noblitt",
    "Liebold",
    "Breeden",
    "Legros",
    "D'Amore",
    "Sprankle",
    "Parrales",
    "Rossetti",
    "Fonte",
    "Morphis",
    "Hackett",
    "Hazelton",
    "Brook",
    "Zoller",
    "Fiorentino",
    "Norling",
    "Brown",
    "Engler",
    "Bay",
    "Emery",
    "Lindamood",
    "Buesing",
    "Pharr",
    "Church",
    "Strosnider",
    "Hintzen",
    "Turcotte",
    "Carrow",
    "Tibbets",
    "Fadel",
    "Britto",
    "Larkin",
    "Garnica",
    "Pelosi",
    "Groover",
    "Hickle",
    "Bremer",
    "Matteo",
    "Hinkle",
    "Knapp",
    "Peele",
    "Asbill",
    "Nosal",
    "Minear",
    "Pritt",
    "Flagler",
    "Simons",
    "Canton",
    "Hermes",
    "Sapp",
    "Villacorta",
    "Sills",
    "Haley",
    "Starns",
    "Ceron",
    "Bains",
    "Ellenwood",
    "Harrell",
    "Lueras",
    "Shaughnessy",
    "Abernathy",
    "Burson",
    "Chism",
    "Hernandes",
    "Donato",
    "Waters",
    "Ries",
    "Osby",
    "Kuo",
    "Mcnemar",
    "Gallaway",
    "Schultz",
    "Leverich",
    "Ipock",
    "Crinklaw",
    "Gipe",
    "Maggio",
    "Koga",
    "Friscia",
    "Everts",
    "Viands",
    "Camp",
    "Mahler",
    "Grantham",
    "Bernett",
    "Turcios",
    "Cronin",
    "Hamlett",
    "Dunson",
    "Bingaman",
    "Auer",
    "Donnelly",
    "Daniel",
    "Gagliano",
    "Veasley",
    "Newhall",
    "Hartzler",
    "Shover",
    "Flewelling",
    "Altenwerth",
    "Bessey",
    "Stoneman",
    "Donadio",
    "Soderquist",
    "Kovach",
    "Schacherer",
    "Hintz",
    "Grizzle",
    "Pfarr",
    "Willms",
    "Woodrow",
    "Myles",
    "Jennison",
    "Strauch",
    "Joy",
    "Salinas",
    "Marvin",
    "Topp",
    "Vautour",
    "Frierson",
    "Scotti",
    "Fifield",
    "Prigge",
    "Strayer",
    "Weston",
    "Lanterman",
    "Mirelez",
    "Medhurst",
    "Link",
    "Smock",
    "Clothier",
    "Lawhon",
    "Shears",
    "Torp",
    "Samms",
    "Dungy",
    "Frase",
    "Schamberger",
    "Boyle",
    "Trull",
    "Gilley",
    "Lossing",
    "Chauncey",
    "Sax",
    "Gibson",
    "Gall",
    "Esser",
    "Aumick",
    "Kenney",
    "Aufderhar",
    "Saltzman",
    "Zwilling",
    "Delapaz",
    "Beckmann",
    "Leech",
    "Horney",
    "Rakow",
    "Yager",
    "Hermiston",
    "Yann",
    "Ray",
    "Kegler",
    "Albin",
    "Ficklin",
    "Schubert",
    "Willaert",
    "Stairs",
    "Lucian",
    "Lakin",
    "Muma",
    "Hosack",
    "Bayer",
    "Underhill",
    "Klenke",
    "Conine",
    "Abbott",
    "Stutes",
    "Kuvalis",
    "Vanhouten",
    "Gregson",
    "Mcgarr",
    "Reich",
    "Lesch",
    "Cleek",
    "Mattison",
    "Yokota",
    "Meints",
    "Bogner",
    "Mummert",
    "Toon",
    "Upton",
    "Clubb",
    "Kilgore",
    "Remo",
    "Knipp",
    "Gaylord",
    "Wess",
    "Immel",
    "Uhlig",
    "Corkery",
    "Lonzo",
    "Nestle",
    "Belmont",
    "Buice",
    "Primeaux",
    "Taul",
    "Yamasaki",
    "Killingsworth",
    "Hice",
    "Yutzy",
    "Buggs",
    "Domenech",
    "Minton",
    "Silva",
    "Prideaux",
    "Huel",
    "Celestin",
    "Boan",
    "Goertz",
    "Treutel",
    "Kibby",
    "Lemus",
    "Zajac",
    "Reilly",
    "Majewski",
    "Brisco",
    "Spiegel",
    "Atkins",
    "Dunker",
    "Stinchcomb",
    "Casares",
    "Brekke",
    "Dunnington",
    "Sample",
    "Koeppel",
    "Harris",
    "Vavra",
    "Hadfield",
    "Byham",
    "McLaughlin",
    "Krieger",
    "Lykes",
    "Kyser",
    "Rouse",
    "Boyer",
    "Speer",
    "Ramsdell",
    "Jetter",
    "Watsica",
    "Severino",
    "Pringle",
    "Cottone",
    "O'Reilly",
    "Earp",
    "Serafin",
    "Ratchford",
    "Roob",
    "Chason",
    "Reitz",
    "Wedeking",
    "Valenzuela",
    "Paramore",
    "Bowerman",
    "Costigan",
    "Alaimo",
    "Buser",
    "Yost",
    "Oceguera",
    "Barton",
    "Constantino",
    "Pollich",
    "Von",
    "Welch",
    "Kautzer",
    "Berman",
    "Homenick",
    "Lachowicz",
    "Corum",
    "Kenley",
    "Augustyn",
    "Rankins",
    "Duron",
    "Walker",
    "Cruickshank",
    "Haberman",
    "Tracey",
    "Connelly",
    "Wheeless",
    "Corvin",
    "Steil",
    "Sanford",
    "Kutch",
    "Beller",
    "Obey",
    "Voll",
    "Hollifield",
    "Southworth",
    "Collado",
    "Hadaway",
    "Sargeant",
    "Almeida",
    "Conkle",
    "Keyes",
    "Duden",
    "Nolan",
    "Zayac",
    "Sing",
    "Hyatt",
    "Graham",
    "Lintner",
    "Spinka",
    "Cieslak",
    "Weitzel",
    "Rawlings",
    "Coffield",
    "Quinby",
    "Kris",
    "Amy",
    "Willman",
    "Mesta",
    "Thiry",
    "Bye",
    "Pendelton",
    "Hanline",
    "Sayre",
    "Stanton",
    "Elder",
    "Powlowski",
    "Soucy",
    "Range",
    "Finnell",
    "Priester",
    "Morgenstern",
    "Osinski",
    "Paredez",
    "Keeler",
    "Lineman",
    "Cauley",
    "Coggin",
    "Lechuga",
    "Huels",
    "Prager",
    "Tew",
    "Harriston",
    "Vancleave",
    "Johnston",
    "Balistreri",
    "Corwin",
    "Walter",
    "Grimm",
    "Montana",
    "Defilippo",
    "Grasso",
    "Jarrard",
    "Furrow",
    "Garrow",
    "Iskra",
    "Stamm",
    "Zilnicki",
    "Labrie",
    "Satterfield",
    "Tabarez",
    "Eldredge",
    "Neubauer",
    "Garraway",
    "Kulas",
    "Sumner",
    "Pleasant",
    "Dunneback",
    "Glover",
    "Daigneault",
    "Casiano",
    "Szabo",
    "Cardwell",
    "Krings",
    "Hand",
    "Hennessee",
    "Morar",
    "Nipper",
    "Mckey",
    "Seger",
    "Kossman",
    "Alls",
    "Bransford",
    "Rinaldo",
    "Bogart",
    "Orn",
    "Casper",
    "Vincelette",
    "Cree",
    "Arnhold",
    "Mosley",
    "Obando",
    "Stiedemann",
    "Mainor",
    "Mendelsohn",
    "Alcorn",
    "Ranger",
    "Harber",
    "Straughter",
    "Lizaola",
    "Spahn",
    "Seigler",
    "Marra",
    "Cann",
    "Buzzard",
    "Sydnor",
    "Pickney",
    "Quitzon",
    "Kihn",
    "Racicot",
    "Stutts",
    "Lind",
    "Abell",
    "Threlkeld",
    "Roder",
    "Greenfelder",
    "Vanduyn",
    "Howlett",
    "DuBuque",
    "Koch",
    "Kizer",
    "Cassin",
    "Aponte",
    "Pulido",
    "Mccool",
    "Brakus",
    "Predovic",
    "Morissette",
    "Leonard",
    "Permenter",
    "Bernier",
    "Leffler",
    "Deckow",
    "Weatherbee",
    "Kunze",
    "Flick",
    "Tiffany",
    "Lebsack",
    "Razo",
    "Pennel",
    "Cormier",
    "Austria",
    "Blanck",
    "Roehrig",
    "Mccardell",
    "Parisian",
    "Angles",
    "Mccaslin",
    "Matta",
    "Buzbee",
    "Aleshire",
    "Boehm",
    "Brwon",
    "Mcmorris",
    "Ellefson",
    "Handley",
    "Kitch",
    "Ernser",
    "Witting",
    "Greaves",
    "Costillo",
    "Hudson",
    "Oatis",
    "Crosier",
    "Luby",
    "Capp",
    "Krajcik",
    "Sholtis",
    "Haggins",
    "Lockett",
    "Kowal",
    "Saraiva",
    "Manrique",
    "Bloomfield",
    "Towry",
    "Felix",
    "Larry",
    "Cudney",
    "Follmer",
    "Youmans",
    "Nardone",
    "Hearne",
    "Coulter",
    "Oddo",
    "Gitlin",
    "Dennie",
    "Gregoire",
    "Drolet",
    "Harwell",
    "Shepherd",
    "Shadduck",
    "Hudnall",
    "Hibbler",
    "Molnar",
    "Mund",
    "Schinner",
    "Clabaugh",
    "Boatman",
    "Lappin",
    "Perone",
    "Calico",
    "Williamson",
    "Turner",
    "Griffiths",
    "Digennaro",
    "Wykoff",
    "Savoy",
    "Schimke",
    "Burch",
    "Helbing",
    "Nicks",
    "Tipping",
    "Primus",
    "Grabert",
    "Abe",
    "Ascher",
    "Morita",
    "Soltis",
    "Gaston",
    "Larue",
    "Markow",
    "Runte",
    "Fee",
    "Spagnolo",
    "Collins",
    "Hutter",
    "Effertz",
    "Mcclean",
    "Fortunato",
    "Hofer",
    "Hu",
    "Mackin",
    "Exner",
    "Messmer",
    "Gurley",
    "Fredenburg",
    "Elling",
    "Eder",
    "Anderson",
    "Fragale",
    "Trieu",
    "Crumpler",
    "Then",
    "Drennen",
    "Havlik",
    "Finnerty",
    "Beier",
    "Ibarra",
    "VonRueden",
    "Balser",
    "Breitenberg",
    "Shedd",
    "Goold",
    "Pankey",
    "Herman",
    "Alsup",
    "Krasner",
    "Scroggins",
    "Hullinger",
    "Hayes",
    "Leuschke",
    "Eynon",
    "Eubanks",
    "Crochet",
    "Moore",
    "Hadden",
    "Mangrum",
    "Vassell",
    "Geraci",
    "Krogh",
    "Maravilla",
    "Waelchi",
    "Chace",
    "Lamore",
    "Runolfsdottir",
    "Malatesta",
    "Hartson",
    "Ates",
    "Crooks",
    "Red",
    "Kosinski",
    "Hepler",
    "Osier",
    "Police",
    "Welk",
    "Ramos",
    "Baier",
    "Blanda",
    "Harn",
    "Messana",
    "Weber",
    "Pulver",
    "Lurry",
    "Dionne",
    "Pan",
    "Geiser",
    "Kwiecien",
    "Pasco",
    "Vasser",
    "Conard",
    "Rieder",
    "Langosh",
    "Kellogg",
    "Herzog",
    "Viera",
    "Funk",
    "Beahan",
    "Furtado",
    "Sang",
    "Hough",
    "Nickell",
    "Jacobo",
    "Leininger",
    "Westmoreland",
    "Considine",
    "Gessner",
    "Feest",
    "Rodas",
    "Mayer",
    "Holscher",
    "Gabriel",
    "Skow",
    "Hunley",
    "Palazzo",
    "Boruff",
    "Hoard",
    "Rippel",
    "Cimini",
    "Densmore",
    "Tann",
    "Boxx",
    "O'Kon",
    "Godin",
    "O'Keefe",
    "Koelpin",
    "Jasinski",
    "Normand",
    "Aispuro",
    "Luque",
    "Hixson",
    "Ward",
    "Zeitler",
    "Plotner",
    "Bruen",
    "Hahn",
    "Mcbay",
    "Leannon",
    "Vaz",
    "Lybrand",
    "Stonecipher",
    "Peluso",
    "Fenstermaker",
    "Schooley",
    "Gabler",
    "Mcelyea",
    "Ver",
    "Ybarbo",
    "Mancil",
    "Shear",
    "Centers",
    "Fludd",
    "Gloria",
    "Cogdill",
    "Steuber",
    "Eriksen",
    "Platts",
    "Kimes",
    "Gayer",
    "Rounds",
    "Hodkiewicz",
    "Dorner",
    "Dagenhart",
    "Fitting",
    "Bruce",
    "Swift",
    "Hipps",
    "Fitch",
    "Schumm",
    "Carn",
    "Vancleve",
    "Dunmire",
    "Seaton",
    "Zerby",
    "Peters",
    "Hornberger",
    "Gordillo",
    "Egli",
    "Silver",
    "Scheetz",
    "Pridmore",
    "Musselwhite",
    "Payer",
    "Brawn",
    "Cao",
    "Fulop",
    "Huckstep",
    "Gilkison",
    "Twiford",
    "Durrah",
    "Pawlik",
    "Hessel",
    "Lamons",
    "Fink",
    "Devenport",
    "Adan",
    "Jelley",
    "Bumbalough",
    "Cummings",
    "Juergens",
    "Damiani",
    "Escalera",
    "Andrews",
    "Imai",
    "Galley",
    "Borges",
    "Hoopes",
    "Cocco",
    "Ocampo",
    "Uzzle",
    "Irick",
    "Kay",
    "Buttner",
    "Botsford",
    "Ardoin",
    "Fisher",
    "Sutherlin",
    "Libbey",
    "Heller",
    "Foulds",
    "Vita",
    "Jolliff",
    "Holladay",
    "Finneran",
    "Ondricka",
    "Joye",
    "Ogg",
    "Okuneva",
    "Sauage",
    "Jorgenson",
    "Dorn",
    "Rodriques",
    "Putman",
    "Kirst",
    "Furlong",
    "Killam",
    "Leanos",
    "Tank",
    "Unrein",
    "Mcgranahan",
    "Giles",
    "Lindsey",
    "Jacobi",
    "Konopelski",
    "Engman",
    "Minjares",
    "Sears",
    "Bodine",
    "Cheeks",
    "Steele",
    "Brenton",
    "Kiley",
    "Pullum",
    "Rames",
    "Rodriguez",
    "Morrisey",
    "Kear",
    "Maclachlan",
    "Hooser",
    "Modica",
    "Hilll",
    "Kollar",
    "Hammill",
    "Shifflett",
    "Leflore",
    "Cheng",
    "Keeling",
    "Everett",
    "Beauchesne",
    "Leppert",
    "Colegrove",
    "Kampf",
    "Strosin",
    "Warr",
    "Mojarro",
    "Lowe",
    "Moen",
    "Ratke",
    "Denesik",
    "Wetzler",
    "Lemke",
    "Hodson",
    "Prosacco",
    "Savarese",
    "Brasch",
    "Pouros",
    "Wuckert",
    "McGlynn",
    "Tramel",
    "Kopec",
    "Deslauriers",
]



import hashlib
from sensestreet.data.unique_first_names import FIRST_NAMES
from sensestreet.data.unique_last_names import LAST_NAMES
import base64
from typing import Tuple


def anonimize_to_hash(login, short=True):
    """
    Generate an anonymized hash representation of a given login.

    Args:
        login (str): The input string (e.g., username or email) to be anonymized.
        short (bool): If True, returns a Base85-encoded hash for compactness.
                      If False, returns a Base64-encoded hash.

    Returns:
        str: The anonymized hash of the login, encoded as Base85 or Base64.
    """
    hash_value = hashlib.sha512(login.encode()).digest()
    if short:
        return base64.b85encode(hash_value).decode()
    return base64.b64encode(hash_value).decode()

def anonimize_to_name(login) -> Tuple[str, str]:
    """
    Generate a deterministic first and last name based on a login hash.

    Args:
        login (str): The input string (e.g., username or email) to be anonymized.

    Returns:
        Tuple[str, str]: A tuple containing the anonymized first name and last name.
                         The names are selected deterministically from pre-defined lists.
    """
    hash_value = hashlib.sha512(login.encode()).hexdigest()

    first_index = int(hash_value[:64], 16) % len(FIRST_NAMES)
    last_index = int(hash_value[64:128], 16) % len(LAST_NAMES)

    first_name = FIRST_NAMES[first_index]
    last_name = LAST_NAMES[last_index]

    return first_name, last_name



from datetime import datetime, timedelta
import jwt  # pip install pyjwt
from cryptography.hazmat.primitives import serialization



class AuthHeaderBuilder:
    __version__ = "1.0.8"

    def __init__(
        self,
        app_id,
        priv_key_path=None,
        pub_key_id="default.pub",
        server_id="default",
        pub_key_path=None,
        extra_headers=None,
    ):
        self.app_id = app_id
        self.server_id = server_id
        self.pub_key_id = pub_key_id
        self.auth_key = (
            self._load_rsa_private_key(priv_key_path) if priv_key_path else None
        )
        self.last_token_meta = None
        self.last_token = None
        self.extra_headers = extra_headers
        self.pub_key = self._get_pub_key(pub_key_path)

    def get_headers(self, is_not_json=False):
        start_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        head = {}
        if self.auth_key:
            token = self._get_token()
            head["Authorization"] = "Bearer {}".format(token)
        if not is_not_json:
            head["Content-Type"] = "application/json"
        if self.pub_key:
            head["X-key"] = self.pub_key
        head["X-client-version"] = str(self.__version__)
        head["Date"] = start_time
        if isinstance(self.extra_headers, dict):
            for key, value in self.extra_headers.items():
                head[key] = value
        return head

    def _get_token(self):
        if self.last_token:
            safe_time = datetime.now() + timedelta(minutes=1)
            if self.last_token_meta["exp"] > safe_time:
                return self.last_token
        self.last_token, self.last_token_meta = self._generate_token(self.auth_key)
        return self.last_token

    def _generate_token(self, private_key):
        token_payload = {
            # JWT claim
            "exp": datetime.utcnow() + timedelta(minutes=15),  # expiry
            "nbf": datetime.utcnow() - timedelta(minutes=5),
            "aud": "api.sensestreet.com",  # audience
            # Custom fields
            "application_id": self.app_id,
            "public_key_id": self.pub_key_id,
            "server_id": self.server_id,
            "server_role": "default",
        }

        token = jwt.encode(token_payload, private_key, algorithm="RS256")
        return token, token_payload

    def _get_pub_key(self, pub_key_path):
        if pub_key_path:
            try:
                return self._load_rsa_public_key(pub_key_path)
            except FileNotFoundError:
                raise FileNotFoundError("Please provide a valid path to the public key")

    @staticmethod
    def _load_rsa_public_key(path):
        """
        Load public key in pem format.
        """
        with open(path, "rb") as f:
            public_key = f.read()
            decoded_key = public_key.decode()
            if decoded_key[-1] == "\n":
                public_key = public_key[:-1]
        return public_key

    @staticmethod
    def _load_rsa_private_key(path):
        """Load private key in pem format"""
        with open(path, "rb") as f:
            private_key = f.read()
        if (
            b"BEGIN OPENSSH PRIVATE" in private_key
            and b"END OPENSSH PRIVATE" in private_key
        ):
            private_key_openssh = serialization.load_ssh_private_key(private_key, None)
            private_key_rsa = private_key_openssh.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
            return private_key_rsa
        elif b"BEGIN RSA PRIVATE" in private_key and b"END RSA PRIVATE" in private_key:
            return private_key
        raise RuntimeError(
            f"Key in '{path}' is not stored in supported format. Must be either RSA (.pem) or OPENSSH format. "
        )


